import socket
import time

from monasca_agent.collector.checks.services_checks import EventType
from monasca_agent.collector.checks.services_checks import ServicesCheck
from monasca_agent.collector.checks.services_checks import Status


class BadConfException(Exception):
    pass


class TCPCheck(ServicesCheck):

    @staticmethod
    def _load_conf(instance):
        # Fetches the conf

        port = instance.get('port', None)
        timeout = float(instance.get('timeout', 10))
        response_time = instance.get('collect_response_time', False)
        socket_type = None
        try:
            port = int(port)
        except Exception:
            raise BadConfException("%s is not a correct port." % str(port))

        try:
            url = instance.get('host', None)
            split = url.split(":")
        except Exception:  # Would be raised if url is not a string
            raise BadConfException("A valid url must be specified")

        # IPv6 address format: 2001:db8:85a3:8d3:1319:8a2e:370:7348
        if len(split) == 8:  # It may then be a IP V6 address, we check that
            for block in split:
                if len(block) != 4:
                    raise BadConfException("%s is not a correct IPv6 address." % url)

            addr = url
            # It's a correct IP V6 address
            socket_type = socket.AF_INET6

        if socket_type is None:
            try:
                addr = socket.gethostbyname(url)
                socket_type = socket.AF_INET
            except Exception:
                raise BadConfException("URL: %s is not a correct IPv4, IPv6 or hostname" % addr)

        return addr, port, socket_type, timeout, response_time

    def _check(self, instance):
        addr, port, socket_type, timeout, response_time = self._load_conf(instance)
        dimensions = self._set_dimensions(None, instance)
        if instance.get('host'):
            dimensions.update({'url': '%s:%s'.format(instance.get('host'), port)})
        start = time.time()
        try:
            self.log.debug("Connecting to %s %s" % (addr, port))
            sock = socket.socket(socket_type)
            try:
                sock.settimeout(timeout)
                sock.connect((addr, port))
            finally:
                sock.close()

        except socket.timeout as e:
            # The connection timed out because it took more time than the specified
            # value in the yaml config file
            length = int((time.time() - start) * 1000)
            self.log.info("%s:%s is DOWN (%s). Connection failed after %s ms" %
                          (addr, port, str(e), length))
            return Status.DOWN, "%s. Connection failed after %s ms" % (str(e), length)

        except socket.error as e:
            length = int((time.time() - start) * 1000)
            if "timed out" in str(e):

                # The connection timed out because it took more time than the system tcp stack allows
                self.log.warning(
                    "The connection timed out because it took more time than the system tcp stack allows. You might want to change this setting to allow longer timeouts")
                self.log.info("System tcp timeout. Assuming that the checked system is down")
                return Status.DOWN, """Socket error: %s.
                 The connection timed out after %s ms because it took more time than the system tcp stack allows.
                 You might want to change this setting to allow longer timeouts""" % (str(e), length)
            else:
                self.log.info("%s:%s is DOWN (%s). Connection failed after %s ms" %
                              (addr, port, str(e), length))
                return Status.DOWN, "%s. Connection failed after %s ms" % (str(e), length)

        except Exception as e:
            length = int((time.time() - start) * 1000)
            self.log.info("%s:%s is DOWN (%s). Connection failed after %s ms" %
                          (addr, port, str(e), length))
            return Status.DOWN, "%s. Connection failed after %s ms" % (str(e), length)

        if response_time:
            self.gauge('network.tcp.response_time', time.time() - start,
                       dimensions=dimensions)

        self.log.debug("%s:%s is UP" % (addr, port))
        return Status.UP, "UP"

    def _create_status_event(self, status, msg, instance):
        # Get the instance settings
        host = instance.get('host', None)
        port = instance.get('port', None)
        name = instance.get('name', None)
        nb_failures = self.statuses[name].count(Status.DOWN)
        nb_tries = len(self.statuses[name])

        # Get a custom message that will be displayed in the event
        custom_message = instance.get('message', "")
        if custom_message:
            custom_message += " \n"

        # Let the possibility to override the source type name
        instance_source_type_name = instance.get('source_type', None)
        if instance_source_type_name is None:
            source_type = "%s.%s" % (ServicesCheck.SOURCE_TYPE_NAME, name)
        else:
            source_type = "%s.%s" % (ServicesCheck.SOURCE_TYPE_NAME, instance_source_type_name)

        # Get the handles you want to notify
        notify = instance.get('notify', self.init_config.get('notify', []))
        notify_message = ""
        if notify:
            notify_list = []
            for handle in notify:
                notify_list.append("@%s" % handle.strip())
            notify_message = " ".join(notify_list) + " \n"

        if status == Status.DOWN:
            title = "[Alert] %s reported that %s is down" % (self.hostname, name)
            alert_type = "error"
            msg = """%s %s %s reported that %s (%s:%s) failed %s time(s) within %s last attempt(s).
                Last error: %s""" % (notify_message,
                                     custom_message,
                                     self.hostname,
                                     name,
                                     host,
                                     port,
                                     nb_failures,
                                     nb_tries,
                                     msg)
            event_type = EventType.DOWN

        else:  # Status is UP
            title = "[Recovered] %s reported that %s is up" % (self.hostname, name)
            alert_type = "success"
            msg = "%s %s %s reported that %s (%s:%s) recovered." % (
                notify_message, custom_message, self.hostname, name, host, port)
            event_type = EventType.UP

        return {
            'timestamp': int(time.time()),
            'event_type': event_type,
            'host': self.hostname,
            'msg_text': msg,
            'msg_title': title,
            'alert_type': alert_type,
            "source_type_name": source_type,
            "event_object": name,
        }



"""Unix system checks.
"""

# stdlib
import logging
import psutil
import re
import subprocess as sp
import sys
import time

# project

import monasca_agent.collector.checks.check as check
import monasca_agent.common.metrics as metrics
import monasca_agent.common.util as util


# locale-resilient float converter
to_float = lambda s: float(s.replace(",", "."))


class Disk(check.Check):
    """Collects metrics about the machine's disks.
    """

    def check(self):
        """Get disk space/inode stats.
        """
        # psutil can be used for disk usage stats but not for inode information, so df is used.

        fs_types_to_ignore = []
        # First get the configuration.
        if self.agent_config is not None:
            use_mount = self.agent_config.get("use_mount", False)
            blacklist_re = self.agent_config.get('device_blacklist_re', None)
            for fs_type in self.agent_config.get('ignore_filesystem_types', []):
                fs_types_to_ignore.extend(['-x', fs_type])
        else:
            use_mount = False
            blacklist_re = None
        platform_name = sys.platform

        try:
            dfk_out = _get_subprocess_output(['df', '-k'] + fs_types_to_ignore)
            stats = self.parse_df_output(
                dfk_out,
                platform_name,
                use_mount=use_mount,
                blacklist_re=blacklist_re
            )

            # Collect inode metrics.
            dfi_out = _get_subprocess_output(['df', '-i'] + fs_types_to_ignore)
            inodes = self.parse_df_output(
                dfi_out,
                platform_name,
                inodes=True,
                use_mount=use_mount,
                blacklist_re=blacklist_re
            )
            # parse into a list of Measurements
            stats.update(inodes)
            timestamp = time.time()
            measurements = [metrics.Measurement(key.split('.', 1)[1],
                                                timestamp,
                                                value,
                                                self._set_dimensions({'device': key.split('.', 1)[0]}),
                                                None)
                            for key, value in stats.iteritems()]

            return measurements

        except Exception:
            self.logger.exception('Error collecting disk stats')
            return []

    def parse_df_output(
            self, df_output, platform_name, inodes=False, use_mount=False, blacklist_re=None):
        """Parse the output of the df command.

        If use_volume is true the volume is used to anchor the metric, otherwise false the mount
        point is used. Returns a tuple of (disk, inode).
        """
        usage_data = {}

        # Transform the raw output into tuples of the df data.
        devices = self._transform_df_output(df_output, blacklist_re)

        # If we want to use the mount point, replace the volume name on each line.
        for parts in devices:
            try:
                if use_mount:
                    parts[0] = parts[-1]
                if inodes:
                    if util.Platform.is_darwin(platform_name):
                        # Filesystem 512-blocks Used Available Capacity iused ifree %iused  Mounted
                        # Inodes are in position 5, 6 and we need to compute the total
                        # Total
                        parts[1] = int(parts[5]) + int(parts[6])  # Total
                        parts[2] = int(parts[5])  # Used
                        parts[3] = int(parts[6])  # Available
                    elif util.Platform.is_freebsd(platform_name):
                        # Filesystem 1K-blocks Used Avail Capacity iused ifree %iused Mounted
                        # Inodes are in position 5, 6 and we need to compute the total
                        parts[1] = int(parts[5]) + int(parts[6])  # Total
                        parts[2] = int(parts[5])  # Used
                        parts[3] = int(parts[6])  # Available
                    else:
                        parts[1] = int(parts[1])  # Total
                        parts[2] = int(parts[2])  # Used
                        parts[3] = int(parts[3])  # Available
                else:
                    parts[1] = int(parts[1])  # Total
                    parts[2] = int(parts[2])  # Used
                    parts[3] = int(parts[3])  # Available
            except IndexError:
                self.logger.exception("Cannot parse %s" % (parts,))

            # Some partitions (EFI boot) may appear to have 0 available inodes
            if parts[1] == 0:
                continue

            #
            # Remote shared storage device names like '10.103.0.220:/instances'
            # cause invalid metrics on the api server side, so if we encounter
            # a colon, remove everything to the left of it (including the
            # offending colon).
            #
            device_name = parts[0]
            idx = device_name.find(":")
            if idx > 0:
                device_name = device_name[(idx+1):]
            if inodes:
                usage_data['%s.disk.inode_used_perc' % device_name] = float(parts[2]) / parts[1] * 100
            else:
                usage_data['%s.disk.space_used_perc' % device_name] = float(parts[2]) / parts[1] * 100

        return usage_data

    @staticmethod
    def _is_number(a_string):
        try:
            float(a_string)
        except ValueError:
            return False
        return True

    def _is_real_device(self, device):
        """Return true if we should track the given device name and false otherwise.
        """
        # First, skip empty lines.
        if not device or len(device) <= 1:
            return False

        # Filter out fake devices.
        device_name = device[0]
        if device_name == 'none':
            return False

        # Now filter our fake hosts like 'map -hosts'. For example:
        #       Filesystem    1024-blocks     Used Available Capacity  Mounted on
        #       /dev/disk0s2    244277768 88767396 155254372    37%    /
        #       map -hosts              0        0         0   100%    /net
        blocks = device[1]
        if not self._is_number(blocks):
            return False
        return True

    def _flatten_devices(self, devices):
        # Some volumes are stored on their own line. Rejoin them here.
        previous = None
        for parts in devices:
            if len(parts) == 1:
                previous = parts[0]
            elif previous and self._is_number(parts[0]):
                # collate with previous line
                parts.insert(0, previous)
                previous = None
            else:
                previous = None
        return devices

    def _transform_df_output(self, df_output, blacklist_re):
        """Given raw output for the df command, transform it into a normalized list devices.

        A 'device' is a list with fields corresponding to the output of df output on each platform.
        """
        all_devices = [l.strip().split() for l in df_output.split("\n")]

        # Skip the header row and empty lines.
        raw_devices = [l for l in all_devices[1:] if l]

        # Flatten the disks that appear in the mulitple lines.
        flattened_devices = self._flatten_devices(raw_devices)

        # Filter fake disks.
        def keep_device(device):
            if not self._is_real_device(device):
                return False
            if blacklist_re and blacklist_re.match(device[0]):
                return False
            return True

        devices = filter(keep_device, flattened_devices)

        return devices


class IO(check.Check):

    def __init__(self, logger, agent_config=None):
        super(IO, self).__init__(logger, agent_config)
        self.header_re = re.compile(r'([%\\/\-_a-zA-Z0-9]+)[\s+]?')
        self.item_re = re.compile(r'^([a-zA-Z0-9\/]+)')
        self.value_re = re.compile(r'\d+\.\d+')
        self.stat_blacklist = ["await", "wrqm/s", "avgqu-sz", "r_await", "w_await", "rrqm/s",
                               "avgrq-sz", "%util", "svctm"]

    def _parse_linux2(self, output):
        recent_stats = output.split('Device:')[2].split('\n')
        header = recent_stats[0]
        header_names = re.findall(self.header_re, header)

        io_stats = {}

        for statsIndex in range(1, len(recent_stats)):
            row = recent_stats[statsIndex]

            if not row:
                # Ignore blank lines.
                continue

            device_match = self.item_re.match(row)

            if device_match is not None:
                # Sometimes device names span two lines.
                device = device_match.groups()[0]
            else:
                continue

            values = re.findall(self.value_re, row)

            if not values:
                # Sometimes values are on the next line so we encounter
                # instances of [].
                continue

            io_stats[device] = {}

            for header_index in range(len(header_names)):
                header_name = header_names[header_index]
                io_stats[device][self.xlate(header_name, "linux")] = values[header_index]

        return io_stats

    @staticmethod
    def _parse_darwin(output):
        lines = [l.split() for l in output.split("\n") if len(l) > 0]
        disks = lines[0]
        lastline = lines[-1]
        io = {}
        for idx, disk in enumerate(disks):
            kb_t, tps, mb_s = map(float, lastline[(3 * idx):(3 * idx) + 3])  # 3 cols at a time
            io[disk] = {
                'system.io.bytes_per_s': mb_s * 10 ** 6,
            }
        return io

    @staticmethod
    def xlate(metric_name, os_name):
        """Standardize on linux metric names.
        """
        if os_name == "sunos":
            names = {"wait": "await",
                     "svc_t": "svctm",
                     "%b": "%util",
                     "kr/s": "io.read_kbytes_sec",
                     "kw/s": "io.write_kbytes_sec",
                     "actv": "avgqu-sz"}
        elif os_name == "freebsd":
            names = {"svc_t": "await",
                     "%b": "%util",
                     "kr/s": "io.read_kbytes_sec",
                     "kw/s": "io.write_kbytes_sec",
                     "wait": "avgqu-sz"}
        elif os_name == "linux":
            names = {"rkB/s": "io.read_kbytes_sec",
                     "r/s": "io.read_req_sec",
                     "wkB/s": "io.write_kbytes_sec",
                     "w/s": "io.write_req_sec"}
        # translate if possible
        return names.get(metric_name, metric_name)

    def check(self):
        """Capture io stats.

        @rtype dict
        @return [metrics.Measurement, ]
        """

        # TODO psutil.disk_io_counters() will also return this infomration but it isn't per second and so must be
        # converted possibly by doing two runs a second apart or storing timestamp+data from the previous collection
        io = {}
        try:
            if util.Platform.is_linux():
                stdout = sp.Popen(['iostat', '-d', '1', '2', '-x', '-k'],
                                  stdout=sp.PIPE,
                                  close_fds=True).communicate()[0]

                #                 Linux 2.6.32-343-ec2 (ip-10-35-95-10)   12/11/2012      _x86_64_        (2 CPU)
                #
                # Device:     rrqm/s   wrqm/s     r/s     w/s    rkB/s    wkB/s avgrq-sz avgqu-sz   await  svctm  %util
                # sda1          0.00    17.61    0.26   32.63     4.23   201.04    12.48     0.16    4.81   0.53   1.73
                # sdb           0.00     2.68    0.19    3.84     5.79    26.07    15.82     0.02    4.93   0.22   0.09
                # sdg           0.00     0.13    2.29    3.84   100.53    30.61    42.78     0.05    8.41   0.88   0.54
                # sdf           0.00     0.13    2.30    3.84   100.54    30.61    42.78     0.06    9.12   0.90   0.55
                # md0           0.00     0.00    0.05    3.37     1.41    30.01    18.35     0.00    0.00   0.00   0.00
                #
                # Device:     rrqm/s   wrqm/s     r/s     w/s    rkB/s    wkB/s avgrq-sz avgqu-sz   await  svctm  %util
                # sda1          0.00     0.00    0.00   10.89     0.00    43.56     8.00     0.03    2.73   2.73   2.97
                # sdb           0.00     0.00    0.00    2.97     0.00    11.88     8.00     0.00    0.00   0.00   0.00
                # sdg           0.00     0.00    0.00    0.00     0.00     0.00     0.00     0.00    0.00   0.00   0.00
                # sdf           0.00     0.00    0.00    0.00     0.00     0.00     0.00     0.00    0.00   0.00   0.00
                # md0           0.00     0.00    0.00    0.00     0.00     0.00     0.00
                # 0.00    0.00   0.00   0.00
                io.update(self._parse_linux2(stdout))

            elif sys.platform == "sunos5":
                iostat = sp.Popen(["iostat", "-x", "-d", "1", "2"],
                                  stdout=sp.PIPE,
                                  close_fds=True).communicate()[0]

                #                   extended device statistics <-- since boot
                # device      r/s    w/s   kr/s   kw/s wait actv  svc_t  %w  %b
                # ramdisk1    0.0    0.0    0.1    0.1  0.0  0.0    0.0   0   0
                # sd0         0.0    0.0    0.0    0.0  0.0  0.0    0.0   0   0
                # sd1        79.9  149.9 1237.6 6737.9  0.0  0.5    2.3   0  11
                #                   extended device statistics <-- past second
                # device      r/s    w/s   kr/s   kw/s wait actv  svc_t  %w  %b
                # ramdisk1    0.0    0.0    0.0    0.0  0.0  0.0    0.0   0   0
                # sd0         0.0    0.0    0.0    0.0  0.0  0.0    0.0   0   0
                # sd1         0.0  139.0    0.0 1850.6  0.0  0.0    0.1   0   1

                # discard the first half of the display (stats since boot)
                lines = [l for l in iostat.split("\n") if len(l) > 0]
                lines = lines[len(lines) / 2:]

                assert "extended device statistics" in lines[0]
                headers = lines[1].split()
                assert "device" in headers
                for l in lines[2:]:
                    cols = l.split()
                    # cols[0] is the device
                    # cols[1:] are the values
                    io[cols[0]] = {}
                    for i in range(1, len(cols)):
                        io[cols[0]][self.xlate(headers[i], "sunos")] = cols[i]

            elif sys.platform.startswith("freebsd"):
                iostat = sp.Popen(["iostat", "-x", "-d", "1", "2"],
                                  stdout=sp.PIPE,
                                  close_fds=True).communicate()[0]

                # Be careful!
                # It looks like SunOS, but some columms (wait, svc_t) have different meaning
                #                        extended device statistics
                # device     r/s   w/s    kr/s    kw/s wait svc_t  %b
                # ad0        3.1   1.3    49.9    18.8    0   0.7   0
                #                         extended device statistics
                # device     r/s   w/s    kr/s    kw/s wait svc_t  %b
                # ad0        0.0   2.0     0.0    31.8    0   0.2   0

                # discard the first half of the display (stats since boot)
                lines = [l for l in iostat.split("\n") if len(l) > 0]
                lines = lines[len(lines) / 2:]

                assert "extended device statistics" in lines[0]
                headers = lines[1].split()
                assert "device" in headers
                for l in lines[2:]:
                    cols = l.split()
                    # cols[0] is the device
                    # cols[1:] are the values
                    io[cols[0]] = {}
                    for i in range(1, len(cols)):
                        io[cols[0]][self.xlate(headers[i], "freebsd")] = cols[i]
            elif sys.platform == 'darwin':
                iostat = sp.Popen(['iostat', '-d', '-c', '2', '-w', '1'],
                                  stdout=sp.PIPE,
                                  close_fds=True).communicate()[0]
                #          disk0           disk1          <-- number of disks
                #    KB/t tps  MB/s     KB/t tps  MB/s
                #   21.11  23  0.47    20.01   0  0.00
                #    6.67   3  0.02     0.00   0  0.00    <-- line of interest
                io = self._parse_darwin(iostat)
            else:
                return []

            # If we filter devices, do it know.
            if self.agent_config is not None:
                device_blacklist_re = self.agent_config.get('device_blacklist_re', None)
            else:
                device_blacklist_re = None
            if device_blacklist_re:
                filtered_io = {}
                for device, stats in io.iteritems():
                    if not device_blacklist_re.match(device):
                        filtered_io[device] = stats
            else:
                filtered_io = io

            measurements = []
            timestamp = time.time()
            for dev_name, stats in filtered_io.iteritems():
                filtered_stats = dict((stat, stats[stat])
                                  for stat in stats.iterkeys() if stat not in self.stat_blacklist)
                m_list = [metrics.Measurement(key,
                                              timestamp,
                                              value,
                                              self._set_dimensions({'device': dev_name}),
                                              None)
                          for key, value in filtered_stats.iteritems()]
                measurements.extend(m_list)

            return measurements

        except Exception:
            self.logger.exception("Cannot extract IO statistics")
            return []


class Load(check.Check):

    def check(self):
        if util.Platform.is_linux():
            try:
                loadAvrgProc = open('/proc/loadavg', 'r')
                uptime = loadAvrgProc.readlines()
                loadAvrgProc.close()
            except Exception:
                self.logger.exception('Cannot extract load')
                return []

            uptime = uptime[0]  # readlines() provides a list but we want a string

        elif sys.platform in ('darwin', 'sunos5') or sys.platform.startswith("freebsd"):
            # Get output from uptime
            try:
                uptime = sp.Popen(['uptime'],
                                  stdout=sp.PIPE,
                                  close_fds=True).communicate()[0]
            except Exception:
                self.logger.exception('Cannot extract load')
                return {}

        # Split out the 3 load average values
        load = [res.replace(',', '.') for res in re.findall(r'([0-9]+[\.,]\d+)', uptime)]
        timestamp = time.time()
        dimensions = self._set_dimensions(None)

        return [metrics.Measurement('load.avg_1_min', timestamp, float(load[0]), dimensions),
                metrics.Measurement('load.avg_5_min', timestamp, float(load[1]), dimensions),
                metrics.Measurement('load.avg_15_min', timestamp, float(load[2]), dimensions)]


class Memory(check.Check):

    def check(self):
        mem_info = psutil.virtual_memory()
        swap_info = psutil.swap_memory()
        mem_data = {
            'mem.total_mb': int(mem_info.total/1048576),
            'mem.free_mb': int(mem_info.free/1048576),
            'mem.usable_mb': int(mem_info.available/1048576),
            'mem.usable_perc': float(100 - mem_info.percent),
            'mem.swap_total_mb': int(swap_info.total/1048576),
            'mem.swap_used_mb': int(swap_info.used/1048576),
            'mem.swap_free_mb': int(swap_info.free/1048576),
            'mem.swap_free_perc': float(100 - swap_info.percent)
        }

        if 'buffers' in mem_info:
            mem_data['mem.used_buffers'] = int(mem_info.buffers/1048576)

        if 'cached' in mem_info:
            mem_data['mem.used_cache'] = int(mem_info.cached/1048576)

        if 'shared' in mem_info:
            mem_data['mem.used_shared'] = int(mem_info.shared/1048576)

        timestamp = time.time()
        dimensions = self._set_dimensions(None)
        return [metrics.Measurement(key, timestamp, value, dimensions) for key, value in mem_data.iteritems()]


class Cpu(check.Check):

    def check(self):
        """Return an aggregate of CPU stats across all CPUs.
        """
        cpu_stats = psutil.cpu_times_percent(percpu=False)
        return self._format_results(cpu_stats.user + cpu_stats.nice,
                                    cpu_stats.system + cpu_stats.irq + cpu_stats.softirq,
                                    cpu_stats.iowait,
                                    cpu_stats.idle,
                                    cpu_stats.steal)

    def _format_results(self, us, sy, wa, idle, st):
        data = {'cpu.user_perc': us,
                'cpu.system_perc': sy,
                'cpu.wait_perc': wa,
                'cpu.idle_perc': idle,
                'cpu.stolen_perc': st}
        for key in data.keys():
            if data[key] is None:
                del data[key]

        timestamp = time.time()
        dimensions = self._set_dimensions(None)
        return [metrics.Measurement(key, timestamp, value, dimensions) for key, value in data.iteritems()]


def _get_subprocess_output(command):
    """Run the given subprocess command and return it's output.

    Raise an Exception if an error occurs.
    """
    proc = sp.Popen(command, stdout=sp.PIPE, close_fds=True)
    return proc.stdout.read()


if __name__ == '__main__':
    # 1s loop with results

    logging.basicConfig(level=logging.DEBUG, format='%(asctime)-15s %(message)s')
    log = logging.getLogger()
    config = {"device_blacklist_re": re.compile('.*disk0.*')}
    cpu = Cpu(log, config)
    disk = Disk(log, config)
    io = IO(log, config)
    load = Load(log, config)
    mem = Memory(log, config)

    while True:
        print("=" * 10)
        print("--- IO ---")
        print(io.check())
        print("--- Disk ---")
        print(disk.check())
        print("--- CPU ---")
        print(cpu.check())
        print("--- Load ---")
        print(load.check())
        print("--- Memory ---")
        print(mem.check())
        print("\n\n\n")
        time.sleep(1)



#!/usr/bin/env python
""" Detect running daemons then configure and start the agent.
"""

import argparse
import logging
import os
import platform
import pwd
import socket
import subprocess
import sys
import yaml

import agent_config
from detection.plugins import DETECTION_PLUGINS
import service.sysv as sysv

from detection.utils import check_output

# List of all detection plugins to run
# Map OS to service type
OS_SERVICE_MAP = {'Linux': sysv.SysV}

log = logging.getLogger(__name__)

# dirname is called twice to get the dir 1 above the location of the script
PREFIX_DIR = os.path.dirname(os.path.dirname(os.path.realpath(sys.argv[0])))


def write_template(template_path, out_path, variables, group):
    """ Write a file using a simple python string template.
        Assumes 640 for the permissions and root:group for ownership.
    :param template_path: Location of the Template to use
    :param out_path: Location of the file to write
    :param variables: dictionary with key/value pairs to use in writing the template
    :return: None
    """
    if not os.path.exists(template_path):
        print("Error no template found at {0}".format(template_path))
        sys.exit(1)
    with open(template_path, 'r') as template:
        with open(out_path, 'w') as conf:
            conf.write(template.read().format(**variables))
    os.chown(out_path, 0, group)
    os.chmod(out_path, 0640)


def main(argv=None):
    parser = argparse.ArgumentParser(description='Detect running daemons then configure and start the agent.',
                                     formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument(
        '-u', '--username', help="Username used for keystone authentication", required=True)
    parser.add_argument(
        '-p', '--password', help="Password used for keystone authentication", required=True)
    parser.add_argument('--keystone_url', help="Keystone url", required=True)
    parser.add_argument('--monasca_url', help="Monasca API url, if not defined the url is pulled from keystone", required=False, default='')
    parser.add_argument('--insecure', help="Set whether certificates are used for Keystone authentication", required=False, default=False)
    parser.add_argument('--project_name', help="Project name for keystone authentication", required=False, default='')
    parser.add_argument('--project_domain_id', help="Project domain id for keystone authentication", required=False, default='')
    parser.add_argument('--project_domain_name', help="Project domain name for keystone authentication", required=False, default='')
    parser.add_argument('--project_id', help="Keystone project id  for keystone authentication", required=False, default='')
    parser.add_argument('--ca_file', help="Sets the path to the ca certs file if using certificates. " +
                                          "Required only if insecure is set to False", required=False, default='')
    parser.add_argument('--check_frequency', help="How often to run metric collection in seconds", type=int, default=15)
    parser.add_argument('--config_dir', help="Configuration directory", default='/etc/monasca/agent')
    parser.add_argument('--dimensions', help="Additional dimensions to set for all metrics. A comma seperated list " +
                                             "of name/value pairs, 'name:value,name2:value2'")
    parser.add_argument('--log_dir', help="monasca-agent log directory", default='/var/log/monasca/agent')
    parser.add_argument('--log_level', help="monasca-agent logging level (ERROR, WARNING, INFO, DEBUG)", required=False, default='INFO')
    parser.add_argument(
        '--template_dir', help="Alternative template directory", default=os.path.join(PREFIX_DIR, 'share/monasca/agent'))
    parser.add_argument('--headless', help="Run in a non-interactive mode", action="store_true")
    parser.add_argument('--overwrite',
                        help="Overwrite existing plugin configuration. " +
                             "The default is to merge. Agent.conf is always overwritten.",
                        action="store_true")
    parser.add_argument('--skip_enable', help="By default the service is enabled, " +
                                              "which requires the script run as root. Set this to skip that step.",
                        action="store_true")
    parser.add_argument('--user', help="User name to run monasca-agent as", default='monasca-agent')
    parser.add_argument('-s', '--service', help="Service this node is associated with, added as a dimension.")
    parser.add_argument('--amplifier', help="Integer for the number of additional measurements to create. " +
                                            "Additional measurements contain the 'amplifier' dimension. " +
                                            "Useful for load testing; not for production use.", default=0, required=False)
    parser.add_argument('-v', '--verbose', help="Verbose Output", action="store_true")
    args = parser.parse_args()

    if args.verbose:
        logging.basicConfig(level=logging.DEBUG, format="%(levelname)s: %(message)s")
    else:
        logging.basicConfig(level=logging.INFO, format="%(levelname)s: %(message)s")

    # Detect os
    detected_os = platform.system()
    if detected_os == 'Linux':
        linux_flavor = platform.linux_distribution()[0]
        if 'Ubuntu' or 'debian' in linux_flavor:
            for package in ['coreutils', 'sysstat']:
                # Check for required dependencies for system checks
                try:
                    output = check_output('dpkg -s {0}'.format(package),
                                          stderr=subprocess.STDOUT,
                                          shell=True)
                except subprocess.CalledProcessError:
                    log.warn("*** {0} package is not installed! ***".format(package) +
                             "\nNOTE: If you do not install the {0} ".format(package) +
                             "package, you will not receive all of the standard " +
                             "operating system type metrics!")
        else:
            pass
    elif detected_os == 'Darwin':
        print("Mac OS is not currently supported by the Monasca Agent")
        sys.exit()
    elif detected_os == 'Windows':
        print("Windows is not currently supported by the Monasca Agent")
        sys.exit()
    else:
        print("{0} is not currently supported by the Monasca Agent".format(detected_os))

    # Service enable, includes setup of users/config directories so must be
    # done before configuration
    agent_service = OS_SERVICE_MAP[detected_os](PREFIX_DIR,
                                                args.config_dir,
                                                args.log_dir,
                                                args.template_dir,
                                                username=args.user)
    if not args.skip_enable:
        agent_service.enable()

    gid = pwd.getpwnam(args.user).pw_gid
    # Write the main agent.conf - Note this is always overwritten
    log.info('Configuring base Agent settings.')
    # Join service in with the dimensions
    if args.service:
        if args.dimensions is None:
            args.dimensions = 'service:' + args.service
        else:
            args.dimensions = ','.join([args.dimensions, 'service:' + args.service])
    write_template(os.path.join(args.template_dir, 'agent.conf.template'),
                   os.path.join(args.config_dir, 'agent.conf'),
                   {'args': args, 'hostname': socket.getfqdn() },
                   gid)

    # Write the supervisor.conf
    write_template(os.path.join(args.template_dir, 'supervisor.conf.template'),
                   os.path.join(args.config_dir, 'supervisor.conf'),
                   {'prefix': PREFIX_DIR, 'log_dir': args.log_dir},
                   gid)

    # Run through detection and config building for the plugins
    plugin_config = agent_config.Plugins()
    for detect_class in DETECTION_PLUGINS:
        detect = detect_class(args.template_dir, args.overwrite)
        if detect.available:
            log.info('Configuring {0}'.format(detect.name))
            new_config = detect.build_config()
            plugin_config.merge(new_config)

        # todo add option to install dependencies

    # Write out the plugin config
    for key, value in plugin_config.iteritems():
        # todo if overwrite is set I should either warn or just delete any config files not in the new config
        # todo add the ability to show a diff before overwriting or merging config
        config_path = os.path.join(args.config_dir, 'conf.d', key + '.yaml')
        # merge old and new config, new has precedence
        if (not args.overwrite) and os.path.exists(config_path):
            with open(config_path, 'r') as config_file:
                old_config = yaml.load(config_file.read())
            if old_config is not None:
                agent_config.deep_merge(old_config, value)
                value = old_config
        with open(config_path, 'w') as config_file:
            os.chmod(config_path, 0o640)
            os.chown(config_path, 0, gid)
            config_file.write(yaml.safe_dump(value, encoding='utf-8', allow_unicode=True))

    # Now that the config is build start the service
    try:
        agent_service.start(restart=True)
    except subprocess.CalledProcessError:
        log.error('The service did not startup correctly see %s' % args.log_dir)


if __name__ == "__main__":
    sys.exit(main())



import collections
import Queue
import threading
import time

import monasca_agent.collector.checks
import monasca_agent.collector.checks.libs.thread_pool


TIMEOUT = 180
DEFAULT_SIZE_POOL = 6
MAX_LOOP_ITERATIONS = 1000
FAILURE = "FAILURE"

up_down = collections.namedtuple('up_down', ['UP', 'DOWN'])
Status = up_down('UP', 'DOWN')
EventType = up_down("servicecheck.state_change.up", "servicecheck.state_change.down")


class ServicesCheck(monasca_agent.collector.checks.AgentCheck):
    SOURCE_TYPE_NAME = 'servicecheck'

    """Services checks inherits from this class.

    This class should never be directly instanciated.

    Work flow:
        The main agent loop will call the check function for each instance for
        each iteration of the loop.
        The check method will make an asynchronous call to the _process method in
        one of the thread initiated in the thread pool created in this class constructor.
        The _process method will call the _check method of the inherited class
        which will perform the actual check.

        The _check method must return a tuple which first element is either
            Status.UP or Status.DOWN.
            The second element is a short error message that will be displayed
            when the service turns down.
    """

    def __init__(self, name, init_config, agent_config, instances):
        monasca_agent.collector.checks.AgentCheck.__init__(self, name, init_config, agent_config, instances)

        # A dictionary to keep track of service statuses
        self.statuses = {}
        self.notified = {}
        self.nb_failures = 0
        self.pool_started = False

    def stop(self):
        self.stop_pool()
        self.pool_started = False

    def start_pool(self):
        # The pool size should be the minimum between the number of instances
        # and the DEFAULT_SIZE_POOL. It can also be overridden by the 'threads_count'
        # parameter in the init_config of the check
        self.log.info("Starting Thread Pool")
        default_size = min(self.instance_count(), DEFAULT_SIZE_POOL)
        self.pool_size = int(self.init_config.get('threads_count', default_size))

        self.pool = monasca_agent.collector.checks.libs.thread_pool.Pool(self.pool_size)

        self.resultsq = Queue.Queue()
        self.jobs_status = {}
        self.pool_started = True

    def stop_pool(self):
        self.log.info("Stopping Thread Pool")
        if self.pool_started:
            self.pool.terminate()
            self.pool.join()
            self.jobs_status.clear()
            assert self.pool.get_nworkers() == 0

    def restart_pool(self):
        self.stop_pool()
        self.start_pool()

    def check(self, instance):
        if not self.pool_started:
            self.start_pool()
        # On Windows the agent runs on multiple threads so we need to have an
        # offset of 5 in case the pool_size is 1
        if threading.activeCount() > 5 * self.pool_size + 5:
            raise Exception("Thread number (%s) is exploding. Skipping this check" %
                            threading.activeCount())
        self._process_results()
        self._clean()
        name = instance.get('name', None)
        if name is None:
            self.log.error('Each service check must have a name')
            return

        if name not in self.jobs_status:
            # A given instance should be processed one at a time
            self.jobs_status[name] = time.time()
            self.pool.apply_async(self._process, args=(instance,))
        else:
            self.log.error("Instance: %s skipped because it's already running." % name)

    def _process(self, instance):
        name = instance.get('name', None)

        try:
            status, msg = self._check(instance)

            result = (status, msg, name, instance)
            # We put the results in the result queue
            self.resultsq.put(result)

        except Exception:
            self.log.exception('Failure in ServiceCheck {}'.format(name))
            result = (FAILURE, FAILURE, FAILURE, FAILURE)
            self.resultsq.put(result)

    def _process_results(self):
        for i in range(MAX_LOOP_ITERATIONS):
            try:
                # We want to fetch the result in a non blocking way
                status, msg, name, queue_instance = self.resultsq.get_nowait()
            except Queue.Empty:
                break

            if status == FAILURE:
                self.nb_failures += 1
                if self.nb_failures >= self.pool_size - 1:
                    self.nb_failures = 0
                    self.restart_pool()
                continue

            event = None

            if name not in self.statuses:
                self.statuses[name] = []

            self.statuses[name].append(status)

            window = int(queue_instance.get('window', 1))

            if window > 256:
                self.log.warning("Maximum window size (256) exceeded, defaulting it to 256")
                window = 256

            threshold = queue_instance.get('threshold', 1)

            if len(self.statuses[name]) > window:
                self.statuses[name].pop(0)

            nb_failures = self.statuses[name].count(Status.DOWN)

            if nb_failures >= threshold:
                if self.notified.get(name, Status.UP) != Status.DOWN:
                    event = self._create_status_event(status, msg, queue_instance)
                    self.notified[name] = Status.DOWN
            else:
                if self.notified.get(name, Status.UP) != Status.UP:
                    event = self._create_status_event(status, msg, queue_instance)
                    self.notified[name] = Status.UP

            if event is not None:
                self.events.append(event)

            # The job is finished here, this instance can be re processed
            del self.jobs_status[name]

    def _check(self, instance):
        """This function should be implemented by inherited classes.
        """
        raise NotImplementedError

    def _clean(self):
        now = time.time()
        for name in self.jobs_status.keys():
            start_time = self.jobs_status[name]
            if now - start_time > TIMEOUT:
                self.log.critical("Restarting Pool. One check is stuck.")
                self.restart_pool()



import unittest
import subprocess
import os
import sys
from random import random, randrange
import urllib as url
import time

sys.path.append(os.path.join(os.path.dirname(__file__), '../monasca_agent'))
from monasca_agent.collector.daemon import CollectorDaemon
from monasca_agent.common.util import Watchdog


class TestWatchdog(unittest.TestCase):

    """Test watchdog in various conditions
    """

    JITTER_FACTOR = 2

    def test_watchdog(self):
        """Verify that watchdog kills ourselves even when spinning
        Verify that watchdog kills ourselves when hanging
        """
        start = time.time()
        try:
            subprocess.check_call(
                ["python", "tests/test_watchdog.py", "busy"], stderr=subprocess.STDOUT)
            raise Exception("Should have died with an error")
        except subprocess.CalledProcessError:
            duration = int(time.time() - start)
            self.assertTrue(duration < self.JITTER_FACTOR * 5)

        # Start pseudo web server
        subprocess.Popen(["nc", "-l", "31834"])
        start = time.time()
        try:
            subprocess.check_call(["python", "tests/test_watchdog.py", "net"])
            raise Exception("Should have died with an error")
        except subprocess.CalledProcessError:
            duration = int(time.time() - start)
            self.assertTrue(duration < self.JITTER_FACTOR * 5)

        # Normal loop, should run 5 times
        start = time.time()
        try:
            subprocess.check_call(["python", "tests/test_watchdog.py", "normal"])
            duration = int(time.time() - start)
            self.assertTrue(duration < self.JITTER_FACTOR * 5)
        except subprocess.CalledProcessError:
            self.fail("Watchdog killed normal process after %s seconds" % int(time.time() - start))

        # Fast tornado, not killed
        start = time.time()
        p = subprocess.Popen(["python", "tests/test_watchdog.py", "fast"])
        p.wait()
        duration = int(time.time() - start)
        # should die as soon as flush_trs has been called
        self.assertTrue(duration, self.JITTER_FACTOR * 10)

        # Slow tornado, killed by the Watchdog
        start = time.time()
        p = subprocess.Popen(["python", "tests/test_watchdog.py", "slow"])
        p.wait()
        duration = int(time.time() - start)
        self.assertTrue(duration < self.JITTER_FACTOR * 4)

        # Too much memory used, killed by Watchdog
        start = time.time()
        p = subprocess.Popen(["python", "tests/test_watchdog.py", "memory"])
        p.wait()
        duration = int(time.time() - start)
        # process should be killed well before the restart interval of 30.
        assert duration < 20


class MockTxManager(object):

    @staticmethod
    def flush():
        "Pretend to flush for a long time"
        time.sleep(5)
        sys.exit(0)


class MemoryHogTxManager(object):

    @staticmethod
    def flush():
        rand_data = []
        while True:
            rand_data.append('%030x' % randrange(256 ** 15))


class PseudoAgent(object):

    """Same logic as the agent, simplified"""
    @staticmethod
    def busy_run():
        w = Watchdog(5)
        w.reset()
        x = 0
        while True:
            x = random()
        return x

    @staticmethod
    def hanging_net():
        w = Watchdog(5)
        w.reset()
        x = url.urlopen("http://localhost:31834")
        print("ERROR Net call returned", x)
        return True

    @staticmethod
    def normal_run():
        w = Watchdog(2)
        w.reset()
        for i in range(5):
            time.sleep(1)
            w.reset()

    @staticmethod
    def slow_tornado():
        a = CollectorDaemon(12345, {})
        a._watchdog = Watchdog(4)
        a._tr_manager = MockTxManager()
        a.run()

    @staticmethod
    def fast_tornado():
        a = CollectorDaemon(12345, {})
        a._watchdog = Watchdog(6)
        a._tr_manager = MockTxManager()
        a.run()

    @staticmethod
    def use_lots_of_memory():
        # Skip this step on travis
        if os.environ.get('TRAVIS', False):
            return
        a = CollectorDaemon(12345, {})
        a._watchdog = Watchdog(30, 50)
        a._tr_manager = MemoryHogTxManager()
        a.run()

if __name__ == "__main__":
    if sys.argv[1] == "busy":
        a = PseudoAgent()
        a.busy_run()
    elif sys.argv[1] == "net":
        a = PseudoAgent()
        a.hanging_net()
    elif sys.argv[1] == "normal":
        a = PseudoAgent()
        a.normal_run()
    elif sys.argv[1] == "slow":
        a = PseudoAgent()
        a.slow_tornado()
    elif sys.argv[1] == "fast":
        a = PseudoAgent()
        a.fast_tornado()
    elif sys.argv[1] == "test":
        t = TestWatchdog()
        t.runTest()
    elif sys.argv[1] == "memory":
        a = PseudoAgent()
        a.use_lots_of_memory()



import logging

from plugin import Plugin

from monasca_setup import agent_config
from monasca_setup.detection import find_process_cmdline
from monasca_setup.detection import service_api_check
from monasca_setup.detection import watch_process


log = logging.getLogger(__name__)


class ServicePlugin(Plugin):

    """Base class implemented by the monasca-agent plugin detection classes

       for OpenStack Services
    """

    def __init__(self, kwargs):
        self.service_name = kwargs['service_name']
        self.process_names = kwargs['process_names']
        self.service_api_url = kwargs['service_api_url']
        self.search_pattern = kwargs['search_pattern']

        super(ServicePlugin, self).__init__(kwargs['template_dir'], kwargs['overwrite'])

    def _detect(self):
        """Run detection.

        """
        self.found_processes = []

        for process in self.process_names:
            if find_process_cmdline(process) is not None:
                self.found_processes.append(process)
        if len(self.found_processes) > 0:
            self.available = True

    def build_config(self):
        """Build the config as a Plugins object and return.

        """
        config = agent_config.Plugins()
        for process in self.found_processes:
            # Watch the service processes
            log.info("\tMonitoring the {0} {1} process.".format(process, self.service_name))
            config.merge(watch_process([process], self.service_name, process, exact_match=False))

        if self.service_api_url and self.search_pattern:
            # Setup an active http_status check on the API
            log.info("\tConfiguring an http_check for the {0} API.".format(self.service_name))
            config.merge(service_api_check(self.service_name + '-api', self.service_api_url,
                                           self.search_pattern, self.service_name))

        return config

    def dependencies_installed(self):
        """Return True if dependencies are installed.

        """
        return True



import monasca_agent.collector.checks.check

try:
    import wmi
    w = wmi.WMI()
except Exception:
    wmi, w = None, None


# Device WMI drive types
class DriveType(object):
    UNKNOWN, NOROOT, REMOVEABLE, LOCAL, NETWORK, CD, RAM = (0, 1, 2, 3, 4, 5, 6)
IGNORED = ('_total',)
B2MB = float(1048576)
KB2MB = B2KB = float(1024)


class Processes(monasca_agent.collector.checks.check.Check):

    def __init__(self, logger):
        monasca_agent.collector.checks.check.Check.__init__(self, logger)
        self.gauge('system.proc.queue_length')
        self.gauge('system.proc.count')

    def check(self, agentConfig):
        try:
            os = w.Win32_PerfFormattedData_PerfOS_System()[0]
        except AttributeError:
            self.logger.info('Missing Win32_PerfFormattedData_PerfOS_System WMI class.' +
                             ' No process metrics will be returned.')
            return

        try:
            w.Win32_PerfFormattedData_PerfOS_Processor(name="_Total")[0]
        except AttributeError:
            self.logger.info('Missing Win32_PerfFormattedData_PerfOS_Processor WMI class.' +
                             ' No process metrics will be returned.')
            return
        if os.ProcessorQueueLength is not None:
            self.save_sample('system.proc.queue_length', os.ProcessorQueueLength)
        if os.Processes is not None:
            self.save_sample('system.proc.count', os.Processes)

        return self.get_metrics()


class Memory(monasca_agent.collector.checks.check.Check):

    def __init__(self, logger):
        monasca_agent.collector.checks.check.Check.__init__(self, logger)
        self.logger = logger
        self.gauge('system.mem.free')
        self.gauge('system.mem.used')
        self.gauge('system.mem.total')
        self.gauge('system.mem.cached')
        self.gauge('system.mem.committed')
        self.gauge('system.mem.paged')
        self.gauge('system.mem.nonpaged')

    def check(self, agentConfig):
        try:
            os = w.Win32_OperatingSystem()[0]
        except AttributeError:
            self.logger.info('Missing Win32_OperatingSystem. No memory metrics will be returned.')
            return

        if os.TotalVisibleMemorySize is not None and os.FreePhysicalMemory is not None:
            total = int(os.TotalVisibleMemorySize) / KB2MB
            free = int(os.FreePhysicalMemory) / KB2MB
            self.save_sample('system.mem.total', total)
            self.save_sample('system.mem.free', free)
            self.save_sample('system.mem.used', total - free)

        mem = w.Win32_PerfFormattedData_PerfOS_Memory()[0]
        if mem.CacheBytes is not None:
            self.save_sample('system.mem.cached', int(mem.CacheBytes) / B2MB)
        if mem.CommittedBytes is not None:
            self.save_sample('system.mem.committed', int(mem.CommittedBytes) / B2MB)
        if mem.PoolPagedBytes is not None:
            self.save_sample('system.mem.paged', int(mem.PoolPagedBytes) / B2MB)
        if mem.PoolNonpagedBytes is not None:
            self.save_sample('system.mem.nonpaged', int(mem.PoolNonpagedBytes) / B2MB)

        return self.get_metrics()


class Cpu(monasca_agent.collector.checks.check.Check):

    def __init__(self, logger):
        monasca_agent.collector.checks.check.Check.__init__(self, logger)
        self.logger = logger
        self.gauge('system.cpu.user')
        self.gauge('system.cpu.idle')
        self.gauge('system.cpu.interrupt')
        self.gauge('system.cpu.system')

    def check(self, agentConfig):
        try:
            cpu = w.Win32_PerfFormattedData_PerfOS_Processor()
        except AttributeError:
            self.logger.info('Missing Win32_PerfFormattedData_PerfOS_Processor WMI class.' +
                             ' No CPU metrics will be returned.')
            return

        cpu_user = self._average_metric(cpu, 'PercentUserTime')
        if cpu_user:
            self.save_sample('system.cpu.user', cpu_user)

        cpu_idle = self._average_metric(cpu, 'PercentIdleTime')
        if cpu_idle:
            self.save_sample('system.cpu.idle', cpu_idle)

        cpu_interrupt = self._average_metric(cpu, 'PercentInterruptTime')
        if cpu_interrupt is not None:
            self.save_sample('system.cpu.interrupt', cpu_interrupt)

        cpu_privileged = self._average_metric(cpu, 'PercentPrivilegedTime')
        if cpu_privileged is not None:
            self.save_sample('system.cpu.system', cpu_privileged)

        return self.get_metrics()

    @staticmethod
    def _average_metric(wmi_class, wmi_prop):
        """Sum all of the values of a metric from a WMI class object.

        Excludes the value for "_Total"
        """
        val = 0
        counter = 0
        for wmi_object in wmi_class:
            if wmi_object.Name == '_Total':
                # Skip the _Total value
                continue

            if getattr(wmi_object, wmi_prop) is not None:
                counter += 1
                val += float(getattr(wmi_object, wmi_prop))

        if counter > 0:
            return val / counter

        return val


class Network(monasca_agent.collector.checks.check.Check):

    def __init__(self, logger):
        monasca_agent.collector.checks.check.Check.__init__(self, logger)
        self.logger = logger
        self.gauge('system.net.bytes_rcvd')
        self.gauge('system.net.bytes_sent')

    def check(self, agentConfig):
        try:
            net = w.Win32_PerfFormattedData_Tcpip_NetworkInterface()
        except AttributeError:
            self.logger.info('Missing Win32_PerfFormattedData_Tcpip_NetworkInterface WMI class.' +
                             ' No network metrics will be returned')
            return

        for iface in net:
            name = self.normalize_device_name(iface.name)
            if iface.BytesReceivedPerSec is not None:
                self.save_sample('system.net.bytes_rcvd', iface.BytesReceivedPerSec,
                                 device_name=name)
            if iface.BytesSentPerSec is not None:
                self.save_sample('system.net.bytes_sent', iface.BytesSentPerSec,
                                 device_name=name)
        return self.get_metrics()


class Disk(monasca_agent.collector.checks.check.Check):

    def __init__(self, logger):
        monasca_agent.collector.checks.check.Check.__init__(self, logger)
        self.logger = logger
        self.gauge('system.disk.free')
        self.gauge('system.disk.total')
        self.gauge('system.disk.in_use')
        self.gauge('system.disk.used')

    def check(self, agentConfig):
        try:
            disk = w.Win32_LogicalDisk()
        except AttributeError:
            self.logger.info('Missing Win32_LogicalDisk WMI class.' +
                             ' No disk metrics will be returned.')
            return

        for device in disk:
            name = self.normalize_device_name(device.name)
            if device.DriveType in (DriveType.CD, DriveType.UNKNOWN) or name in IGNORED:
                continue
            if device.FreeSpace is not None and device.Size is not None:
                free = float(device.FreeSpace) / B2KB
                total = float(device.Size) / B2KB
                used = total - free
                self.save_sample('system.disk.free', free, device_name=name)
                self.save_sample('system.disk.total', total, device_name=name)
                self.save_sample('system.disk.used', used, device_name=name)
                self.save_sample('system.disk.in_use', (used / total),
                                 device_name=name)
        return self.get_metrics()


class IO(monasca_agent.collector.checks.check.Check):

    def __init__(self, logger):
        monasca_agent.collector.checks.check.Check.__init__(self, logger)
        self.logger = logger
        self.gauge('system.io.wkb_s')
        self.gauge('system.io.w_s')
        self.gauge('system.io.rkb_s')
        self.gauge('system.io.r_s')
        self.gauge('system.io.avg_q_sz')

    def check(self, agentConfig):
        try:
            disk = w.Win32_PerfFormattedData_PerfDisk_LogicalDisk()
        except AttributeError:
            self.logger.info(
                'Missing Win32_PerfFormattedData_PerfDisk_LogicalDiskUnable WMI class.' +
                ' No I/O metrics will be returned.')
            return

        for device in disk:
            name = self.normalize_device_name(device.name)
            if name in IGNORED:
                continue
            if device.DiskWriteBytesPerSec is not None:
                self.save_sample('system.io.wkb_s', int(device.DiskWriteBytesPerSec) / B2KB,
                                 device_name=name)
            if device.DiskWritesPerSec is not None:
                self.save_sample('system.io.w_s', int(device.DiskWritesPerSec),
                                 device_name=name)
            if device.DiskReadBytesPerSec is not None:
                self.save_sample('system.io.rkb_s', int(device.DiskReadBytesPerSec) / B2KB,
                                 device_name=name)
            if device.DiskReadsPerSec is not None:
                self.save_sample('system.io.r_s', int(device.DiskReadsPerSec),
                                 device_name=name)
            if device.CurrentDiskQueueLength is not None:
                self.save_sample('system.io.avg_q_sz', device.CurrentDiskQueueLength,
                                 device_name=name)
        return self.get_metrics()



from collections import defaultdict
import re
import urllib2

from monasca_agent.collector.checks import AgentCheck


db_stats = re.compile(r'^db_(\d)+$')
whitespace = re.compile(r'\s')


class KyotoTycoonCheck(AgentCheck):

    """Report statistics about the Kyoto Tycoon DBM-style

    database server (http://fallabs.com/kyototycoon/)
    """

    GAUGES = {
        'repl_delay': 'replication.delay',
        'serv_thread_count': 'threads',
    }

    RATES = {
        'serv_conn_count': 'connections',
        'cnt_get': 'ops.get.hits',
        'cnt_get_misses': 'ops.get.misses',
        'cnt_set': 'ops.set.hits',
        'cnt_set_misses': 'ops.set.misses',
        'cnt_remove': 'ops.del.hits',
        'cnt_remove_misses': 'ops.del.misses',
    }

    DB_GAUGES = {
        'count': 'records',
        'size': 'size',
    }
    TOTALS = {
        'cnt_get': 'ops.get.total',
        'cnt_get_misses': 'ops.get.total',
        'cnt_set': 'ops.set.total',
        'cnt_set_misses': 'ops.set.total',
        'cnt_remove': 'ops.del.total',
        'cnt_remove_misses': 'ops.del.total',
    }

    def check(self, instance):
        url = instance.get('report_url')
        if not url:
            raise Exception('Invalid Kyoto Tycoon report url %r' % url)

        dimensions = self._set_dimensions(None, instance)
        name = instance.get('name')

        if name is not None:
            dimensions.update({'instance': name})

        response = urllib2.urlopen(url)
        body = response.read()

        totals = defaultdict(lambda: 0)
        for line in body.split('\n'):
            if '\t' not in line:
                continue

            key, value = line.strip().split('\t', 1)
            if key in self.GAUGES:
                name = self.GAUGES[key]
                self.gauge('kyototycoon.%s' % name, float(value), dimensions=dimensions)

            elif key in self.RATES:
                name = self.RATES[key]
                self.rate('kyototycoon.%s_per_s' % name, float(value), dimensions=dimensions)

            elif db_stats.match(key):
                # Also produce a per-db metrics tagged with the db
                # number in addition to the default dimensions
                m = db_stats.match(key)
                dbnum = int(m.group(1))
                db_dimensions = dimensions.copy()
                db_dimensions.update({'db': dbnum})
                for part in whitespace.split(value):
                    k, v = part.split('=', 1)
                    if k in self.DB_GAUGES:
                        name = self.DB_GAUGES[k]
                        self.gauge('kyototycoon.%s' % name, float(v), dimensions=db_dimensions)

            if key in self.TOTALS:
                totals[self.TOTALS[key]] += float(value)

        for key, value in totals.items():
            self.rate('kyototycoon.%s_per_s' % key, value, dimensions=dimensions)



import traceback


def shell():
    from config import get_version

    print("""
Datadog Agent v%s - Python Shell

    """ % (get_version()))
    while True:
        cmd = raw_input('>>> ')
        try:
            exec(cmd)
        except Exception as e:
            print(traceback.format_exc(e))

if __name__ == "__main__":
    shell()



import unittest
from tests.common import load_check

from nose.plugins.attrib import attr
from nose.plugins.skip import SkipTest


class CouchbaseTestCase(unittest.TestCase):

    def setUp(self):
        self.config = {
            'instances': [{
                'server': 'http://localhost:8091',
                'user': 'Administrator',
                'password': 'password',
            }]
        }
        self.agent_config = {
            'version': '0.1',
            'api_key': 'toto'
        }
        self.check = load_check('couchbase', self.config, self.agent_config)

    @attr('couchbase')
    def test_camel_case_to_joined_lower(self):
        test_pairs = {
            'camelCase': 'camel_case',
            'FirstCapital': 'first_capital',
            'joined_lower': 'joined_lower',
            'joined_Upper1': 'joined_upper1',
            'Joined_upper2': 'joined_upper2',
            'Joined_Upper3': 'joined_upper3',
            '_leading_Underscore': 'leading_underscore',
            'Trailing_Underscore_': 'trailing_underscore',
            'DOubleCAps': 'd_ouble_c_aps',
            '@@@super--$$-Funky__$__$$%': 'super_funky',
        }

        for test_input, expected_output in test_pairs.items():
            test_output = self.check.camel_case_to_joined_lower(test_input)
            self.assertEqual(
                test_output,
                expected_output,
                'Input was %s, expected output was %s, actual output was %s' %
                (test_input,
                 expected_output,
                 test_output))

    @attr('couchbase')
    def test_metrics_casing(self):
        raise SkipTest("Skipped for now as it's hard to configure couchbase on travis")
        self.check.check(self.config['instances'][0])

        metrics = self.check.get_metrics()

        camel_cased_metrics = [u'couchbase.hdd.used_by_data',
                               u'couchbase.ram.used_by_data',
                               u'couchbase.ram.quota_total',
                               u'couchbase.ram.quota_used',
                               ]

        found_metrics = [k[0] for k in metrics if k[0] in camel_cased_metrics]
        self.assertEqual(found_metrics.sort(), camel_cased_metrics.sort())

    @attr('couchbase')
    def test_metrics(self):
        raise SkipTest("Skipped for now as it's hard to configure couchbase on travis")
        self.check.check(self.config['instances'][0])

        metrics = self.check.get_metrics()

        self.assertTrue(isinstance(metrics, list), metrics)
        self.assertTrue(len(metrics) > 3)
        self.assertTrue(
            len([k for k in metrics if "instance:http://localhost:8091" in k[3]['dimensions']]) > 3)

        self.assertTrue(len([k for k in metrics if -1 != k[0].find('by_node')])
                        > 1, 'Unable to fund any per node metrics')
        self.assertTrue(len([k for k in metrics if -1 != k[0].find('by_bucket')])
                        > 1, 'Unable to fund any per node metrics')



import sys

if sys.version_info < (2, 6):
    # Normally we'd write our checks to be compatible with >= python 2.4 but
    # the dependencies of this check are not compatible with 2.4 and would
    # be too much work to rewrite, so raise an exception here.
    raise Exception('kafka_consumer check requires at least Python 2.6')

import collections
import monasca_agent.collector.checks as checks

try:
    import kafka.client as client
    import kafka.common as common
    import kafka.consumer as consumer
except ImportError:
    raise Exception('Missing python dependency: kafka (https://github.com/mumrah/kafka-python)')


class KafkaCheck(checks.AgentCheck):
    """ Checks the configured kafka instance reporting the consumption lag
        for each partition per topic in each consumer group. If full_output
        is set also reports broker offsets and the current consumer offset.
        Works on Kafka version >= 0.8.1.1
    """
    def _validate_consumer_groups(self, val):
        """ Private config validation/marshalling functions
        """
        try:
            consumer_group, topic_partitions = val.items()[0]
            assert isinstance(consumer_group, (str, unicode))
            topic, partitions = topic_partitions.items()[0]
            assert isinstance(topic, (str, unicode))
            assert isinstance(partitions, (list, tuple))
            return val
        except Exception as e:
            self.log.exception(e)
            raise Exception('''The `consumer_groups` value must be a mapping of mappings, like this:
consumer_groups:
  myconsumer0: # consumer group name
    mytopic0: [0, 1] # topic: list of partitions
  myconsumer1:
    mytopic0: [0, 1, 2]
    mytopic1: [10, 12]
''')

    def check(self, instance):
        consumer_groups = self.read_config(instance, 'consumer_groups',
                                           cast=self._validate_consumer_groups)
        kafka_host_ports = self.read_config(instance, 'kafka_connect_str')
        full_output = self.read_config(instance, 'full_output', cast=bool)
        dimensions = {'component': 'kafka', 'service': 'kafka'}

        try:
            # Connect to Kafka
            kafka_conn = client.KafkaClient(kafka_host_ports)

            # Query Kafka for consumer offsets
            consumer_offsets = {}
            topics = collections.defaultdict(set)
            for consumer_group, topic_partitions in consumer_groups.iteritems():
                for topic, partitions in topic_partitions.iteritems():
                    kafka_consumer = consumer.SimpleConsumer(kafka_conn, consumer_group, topic)
                    # Remember the topic partitions that we've see so that we can
                    # look up their broker offsets later
                    topics[topic].update(set(partitions))
                    for partition in partitions:
                        try:
                            consumer_offsets[(consumer_group, topic, partition)] = kafka_consumer.offsets[partition]
                        except KeyError:
                            kafka_consumer.stop()
                            self.log.error('Error fetching consumer offset for {} partition {}'.format(topic, partition))
                    kafka_consumer.stop()

            # Query Kafka for the broker offsets, done in a separate loop so only one query is done
            # per topic even if multiple consumer groups watch the same topic
            broker_offsets = {}
            for topic, partitions in topics.items():
                offset_responses = []
                for p in partitions:
                    try:
                        response = kafka_conn.send_offset_request([common.OffsetRequest(topic, p, -1, 1)])
                        offset_responses.append(response[0])
                    except common.KafkaError as e:
                        self.log.error("Error fetching broker offset: {}".format(e))

                for resp in offset_responses:
                    broker_offsets[(resp.topic, resp.partition)] = resp.offsets[0]
        finally:
            try:
                kafka_conn.close()
            except Exception:
                self.log.exception('Error cleaning up Kafka connection')

        # Report the broker data
        if full_output:
            broker_dimensions = dimensions.copy()
            for (topic, partition), broker_offset in broker_offsets.items():
                broker_dimensions.update({'topic': topic,
                                          'partition': partition})
                broker_offset = broker_offsets.get((topic, partition))
                self.gauge('kafka.broker_offset',
                           broker_offset,
                           dimensions=self._set_dimensions(broker_dimensions, instance))

        # Report the consumer data
        consumer_dimensions=dimensions.copy()
        for (consumer_group, topic, partition), consumer_offset in consumer_offsets.items():
            # Get the broker offset
            broker_offset = broker_offsets.get((topic, partition))
            # Report the consumer offset and lag
            consumer_dimensions.update({'topic': topic,
                                        'partition': partition,
                                        'consumer_group': consumer_group})
            if full_output:
                self.gauge('kafka.consumer_offset',
                           consumer_offset,
                           dimensions=self._set_dimensions(consumer_dimensions, instance))
            self.gauge('kafka.consumer_lag',
                       broker_offset - consumer_offset,
                       dimensions=self._set_dimensions(consumer_dimensions, instance))



from fnmatch import fnmatch
from os import stat
from os import walk
from os.path import abspath
from os.path import exists
from os.path import join
import time

from monasca_agent.collector.checks import AgentCheck


class DirectoryCheck(AgentCheck):

    """This check is for monitoring and reporting metrics on the files for a provided directory

    WARNING: the user/group that dd-agent runs as must have access to stat the files in the desired directory

    Config options:
        "directory" - string, the directory to gather stats for. required
        "name" - string, the name to use when tagging the metrics. defaults to the "directory"
        "pattern" - string, the `fnmatch` pattern to use when reading the "directory"'s files. default "*"
        "recursive" - boolean, when true the stats will recurse into directories. default False
    """

    def check(self, instance):
        if "directory" not in instance:
            raise Exception('DirectoryCheck: missing "directory" in config')

        directory = instance["directory"]
        abs_directory = abspath(directory)
        name = instance.get("name") or directory
        pattern = instance.get("pattern") or "*"
        recursive = instance.get("recursive") or False

        if not exists(abs_directory):
            raise Exception("DirectoryCheck: the directory (%s) does not exist" % abs_directory)

        dimensions = self._set_dimensions({"name": name}, instance)
        self._get_stats(abs_directory, name, pattern, recursive, dimensions)

    def _get_stats(self, directory, name, pattern, recursive, dimensions):
        directory_bytes = 0
        directory_files = 0
        for root, dirs, files in walk(directory):
            for filename in files:
                filename = join(root, filename)
                # check if it passes our filter
                if not fnmatch(filename, pattern):
                    continue
                try:
                    file_stat = stat(filename)

                except OSError as ose:
                    self.warning("DirectoryCheck: could not stat file %s - %s" % (filename, ose))
                else:
                    directory_files += 1
                    directory_bytes += file_stat.st_size
                    # file specific metrics
                    self.histogram(
                        "system.disk.directory.file.bytes",
                        file_stat.st_size,
                        dimensions=dimensions)
                    self.histogram(
                        "system.disk.directory.file.modified_sec_ago",
                        time.time() -
                        file_stat.st_mtime,
                        dimensions=dimensions)
                    self.histogram(
                        "system.disk.directory.file.created_sec_ago",
                        time.time() -
                        file_stat.st_ctime,
                        dimensions=dimensions)

            # os.walk gives us all sub-directories and their files
            # if we do not want to do this recursively and just want
            # the top level directory we gave it, then break
            if not recursive:
                break

        # number of files
        self.gauge("system.disk.directory.files", directory_files, dimensions=dimensions)
        # total file size
        self.gauge("system.disk.directory.bytes", directory_bytes, dimensions=dimensions)



"""Classes for detection of running resources to be monitored.

    Detection classes should be platform independent
"""


class Plugin(object):

    """Abstract class implemented by the monasca-agent plugin detection classes.

    """
    # todo these should include dependency detection

    def __init__(self, template_dir, overwrite=True, alarms=None):
        self.available = False
        self.template_dir = template_dir
        self.dependencies = ()
        self.overwrite = overwrite
        self.alarms = alarms
        self._detect()

    def _detect(self):
        """Run detection, set self.available True if the service is detected.

        """
        raise NotImplementedError

    def build_config(self):
        """Build the config as a Plugins object and return.

        """
        raise NotImplementedError

    def dependencies_installed(self):
        """Return True if dependencies are installed.

        """
        raise NotImplementedError

    @property
    def name(self):
        """Return _name if set otherwise the class name.

        """
        if '_name' in self.__dict__:
            return self._name
        else:
            return self.__class__.__name__



# stdlib
import unittest
import logging
import gc
import sys

# 3p
# from nose.plugins.attrib import attr

# project
from monasca_agent.collector import checks as w32
from nose.plugins.skip import SkipTest

logging.basicConfig(level=logging.DEBUG)
log = logging.getLogger(__file__)


AGENT_CONFIG = {}  # None of the windows checks use this.


class TestWin32(unittest.TestCase):

    def _checkMemoryLeak(self, func):
        # FIXME: This should use @attr('windows')instead of checking for the
        # platform, but just importing nose.plugins.attrib causes all the tests
        # to fail with uncollected garbage.
        if sys.platform != 'win32':
            return
        gc.set_debug(gc.DEBUG_LEAK)
        try:
            start = len(gc.garbage)
            func()
            end = len(gc.garbage)
            self.assertEqual(end - start, 0, gc.garbage)
        finally:
            gc.set_debug(0)

    def testDisk(self):
        raise SkipTest("Skipped windows support is temporarily broken.")
        dsk = w32.Disk(log)
        self._checkMemoryLeak(lambda: dsk.check(AGENT_CONFIG))

    def testIO(self):
        raise SkipTest("Skipped windows support is temporarily broken.")
        io = w32.IO(log)
        self._checkMemoryLeak(lambda: io.check(AGENT_CONFIG))

    def testProcesses(self):
        raise SkipTest("Skipped windows support is temporarily broken.")
        proc = w32.Processes(log)
        self._checkMemoryLeak(lambda: proc.check(AGENT_CONFIG))

    def testMemory(self):
        raise SkipTest("Skipped windows support is temporarily broken.")
        mem = w32.Memory(log)
        self._checkMemoryLeak(lambda: mem.check(AGENT_CONFIG))

    def testNetwork(self):
        raise SkipTest("Skipped windows support is temporarily broken.")
        net = w32.Network(log)
        self._checkMemoryLeak(lambda: net.check(AGENT_CONFIG))

    def testCPU(self):
        raise SkipTest("Skipped windows support is temporarily broken.")
        cpu = w32.Cpu(log)
        self._checkMemoryLeak(lambda: cpu.check(AGENT_CONFIG))


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python

# Core modules
import glob
import logging
import os
import signal
import sys
import time

# Custom modules
import checks.collector
import checks.services_checks as status_checks
import jmxfetch
import monasca_agent.common.check_status
import monasca_agent.common.config as cfg
import monasca_agent.common.daemon
import monasca_agent.common.emitter
import monasca_agent.common.util as util

# set up logging before importing any other components
util.initialize_logging('collector')
os.umask(022)

# Check we're not using an old version of Python. We need 2.4 above because
# some modules (like subprocess) were only introduced in 2.4.
if int(sys.version_info[1]) <= 3:
    sys.stderr.write("Monasca Agent requires python 2.4 or later.\n")
    sys.exit(2)

# Constants
PID_NAME = "monasca-agent"
WATCHDOG_MULTIPLIER = 10
RESTART_INTERVAL = 4 * 24 * 60 * 60  # Defaults to 4 days
START_COMMANDS = ['start', 'restart', 'foreground']

# Globals
log = logging.getLogger('collector')


# todo the collector has daemon code but is always run in foreground mode
# from the supervisor, is there a reason for the daemon code then?
class CollectorDaemon(monasca_agent.common.daemon.Daemon):

    """The agent class is a daemon that runs the collector in a background process.

    """

    def __init__(self, pidfile, autorestart, start_event=True):
        monasca_agent.common.daemon.Daemon.__init__(self, pidfile, autorestart=autorestart)
        self.run_forever = True
        self.collector = None
        self.start_event = start_event

    def _handle_sigterm(self, signum, frame):
        log.debug("Caught sigterm. Stopping run loop.")
        self.run_forever = False

        if jmxfetch.JMXFetch.is_running():
            jmxfetch.JMXFetch.stop()

        if self.collector:
            self.collector.stop()
        log.debug("Collector is stopped.")

    def _handle_sigusr1(self, signum, frame):
        self._handle_sigterm(signum, frame)
        self._do_restart()

    def info(self, verbose=None):
        logging.getLogger().setLevel(logging.ERROR)
        return monasca_agent.common.check_status.CollectorStatus.print_latest_status(verbose=verbose)

    def run(self, config):
        """Main loop of the collector.

        """
        # Gracefully exit on sigterm.
        signal.signal(signal.SIGTERM, self._handle_sigterm)

        # A SIGUSR1 signals an exit with an autorestart
        signal.signal(signal.SIGUSR1, self._handle_sigusr1)

        # Handle Keyboard Interrupt
        signal.signal(signal.SIGINT, self._handle_sigterm)

        # Save the agent start-up stats.
        monasca_agent.common.check_status.CollectorStatus().persist()

        # Load the checks_d checks
        checksd = util.load_check_directory()

        self.collector = checks.collector.Collector(config, monasca_agent.common.emitter.http_emitter, checksd)

        # Configure the watchdog.
        check_frequency = int(config['check_freq'])
        watchdog = self._get_watchdog(check_frequency, config)

        # Initialize the auto-restarter
        self.restart_interval = int(config.get('restart_interval', RESTART_INTERVAL))
        self.agent_start = time.time()

        # Run the main loop.
        while self.run_forever:

            # enable profiler if needed
            profiled = False
            if config.get('profile', False) and config.get('profile').lower() == 'yes':
                try:
                    import cProfile
                    profiler = cProfile.Profile()
                    profiled = True
                    profiler.enable()
                    log.debug("Agent profiling is enabled")
                except Exception:
                    log.warn("Cannot enable profiler")

            # Do the work.
            self.collector.run()

            # disable profiler and printout stats to stdout
            if config.get('profile', False) and config.get('profile').lower() == 'yes' and profiled:
                try:
                    profiler.disable()
                    import cStringIO
                    import pstats
                    s = cStringIO.StringIO()
                    ps = pstats.Stats(profiler, stream=s).sort_stats("cumulative")
                    ps.print_stats()
                    log.debug(s.getvalue())
                except Exception:
                    log.warn("Cannot disable profiler")

            # Check if we should restart.
            if self.autorestart and self._should_restart():
                self._do_restart()

            # Only plan for the next loop if we will continue,
            # otherwise just exit quickly.
            if self.run_forever:
                if watchdog:
                    watchdog.reset()
                time.sleep(check_frequency)

        # Now clean-up.
        try:
            monasca_agent.common.check_status.CollectorStatus.remove_latest_status()
        except Exception:
            pass

        # Explicitly kill the process, because it might be running
        # as a daemon.
        log.info("Exiting. Bye bye.")
        sys.exit(0)

    @staticmethod
    def _get_watchdog(check_freq, agentConfig):
        watchdog = None
        if agentConfig.get("watchdog", True):
            watchdog = util.Watchdog(check_freq * WATCHDOG_MULTIPLIER,
                                     max_mem_mb=agentConfig.get('limit_memory_consumption',
                                                                None))
            watchdog.reset()
        return watchdog

    def _should_restart(self):
        if time.time() - self.agent_start > self.restart_interval:
            return True
        return False

    def _do_restart(self):
        log.info("Running an auto-restart.")
        if self.collector:
            self.collector.stop()
        sys.exit(monasca_agent.common.daemon.AgentSupervisor.RESTART_EXIT_STATUS)


def main():
    options, args = util.get_parsed_args()
    config = cfg.Config()
    collector_config = config.get_config(['Main', 'Api', 'Logging'])
    # todo autorestart isn't used remove
    autorestart = collector_config.get('autorestart', False)

    COMMANDS = [
        'start',
        'stop',
        'restart',
        'foreground',
        'status',
        'info',
        'check',
        'check_all',
        'configcheck',
        'jmx',
    ]

    if len(args) < 1:
        sys.stderr.write("Usage: %s %s\n" % (sys.argv[0], "|".join(COMMANDS)))
        return 2

    command = args[0]
    if command not in COMMANDS:
        sys.stderr.write("Unknown command: %s\n" % command)
        return 3

    pid_file = util.PidFile('monasca-agent')

    if options.clean:
        pid_file.clean()

    agent = CollectorDaemon(pid_file.get_path(), autorestart)

    if command in START_COMMANDS:
        log.info('Agent version %s' % config.get_version())

    if 'start' == command:
        log.info('Start daemon')
        agent.start()

    elif 'stop' == command:
        log.info('Stop daemon')
        agent.stop()

    elif 'restart' == command:
        log.info('Restart daemon')
        agent.restart()

    elif 'status' == command:
        agent.status()

    elif 'info' == command:
        return agent.info(verbose=options.verbose)

    elif 'foreground' == command:
        logging.info('Running in foreground')
        if autorestart:
            # Set-up the supervisor callbacks and fork it.
            logging.info('Running Agent with auto-restart ON')

            def child_func():
                agent.run()

            def parent_func():
                agent.start_event = False

            monasca_agent.common.daemon.AgentSupervisor.start(parent_func, child_func)
        else:
            # Run in the standard foreground.
            agent.run(collector_config)

    elif 'check' == command:
        check_name = args[1]
        checks = util.load_check_directory()
        for check in checks['initialized_checks']:
            if check.name == check_name:
                run_check(check)

    elif 'check_all' == command:
        print("Loading check directory...")
        checks = util.load_check_directory()
        print("...directory loaded.\n")
        for check in checks['initialized_checks']:
            run_check(check)

    elif 'configcheck' == command or 'configtest' == command:
        osname = util.get_os()
        all_valid = True
        paths = util.Paths()
        for conf_path in glob.glob(os.path.join(paths.get_confd_path(), "*.yaml")):
            basename = os.path.basename(conf_path)
            try:
                config.check_yaml(conf_path)
            except Exception as e:
                all_valid = False
                print("%s contains errors:\n    %s" % (basename, e))
            else:
                print("%s is valid" % basename)
        if all_valid:
            print("All yaml files passed. You can now run the Monitoring agent.")
            return 0
        else:
            print("Fix the invalid yaml files above in order to start the Monitoring agent. "
                  "A useful external tool for yaml parsing can be found at "
                  "http://yaml-online-parser.appspot.com/")
            return 1

    elif 'jmx' == command:

        if len(args) < 2 or args[1] not in jmxfetch.JMX_LIST_COMMANDS.keys():
            print("#" * 80)
            print("JMX tool to be used to help configure your JMX checks.")
            print("See http://docs.datadoghq.com/integrations/java/ for more information")
            print("#" * 80)
            print("\n")
            print("You have to specify one of the following commands:")
            for command, desc in jmxfetch.JMX_LIST_COMMANDS.iteritems():
                print("      - %s [OPTIONAL: LIST OF CHECKS]: %s" % (command, desc))
            print("Example: sudo /etc/init.d/monasca-agent jmx list_matching_attributes tomcat jmx solr")
            print("\n")

        else:
            jmx_command = args[1]
            checks_list = args[2:]
            paths = util.Paths()
            confd_path = paths.get_confd_path()
            # Start JMXFetch if needed
            should_run = jmxfetch.JMXFetch.init(confd_path,
                                                config,
                                                15,
                                                jmx_command,
                                                checks_list,
                                                reporter="console")
            if not should_run:
                print("Couldn't find any valid JMX configuration in your conf.d directory: %s" % confd_path)
                print("Have you enabled any JMX checks ?")

    return 0

def run_check(check):

    is_multi_threaded = False
    if isinstance(check, status_checks.ServicesCheck):
        is_multi_threaded = True
    print("#" * 80)
    print("Check name: '{}'\n".format(check.name))
    check.run()
    # Sleep for a second and then run a second check to capture rate metrics
    time.sleep(1)
    check.run()
    if is_multi_threaded:
        # Sleep for a second to allow async threads to finish
        time.sleep(1)
        check.stop_pool()
    print("Metrics: ")
    check.get_metrics(prettyprint=True)
    print("#" * 80 + "\n\n")

if __name__ == '__main__':
    try:
        sys.exit(main())
    except Exception:
        # Try our best to log the error.
        try:
            log.exception("Uncaught error running the Agent")
        except Exception:
            pass
        raise



import ConfigParser as parser
import logging
import os
import pkg_resources
import re
import six
import string
import cStringIO as cstringio
import yaml

try:
    from yaml import CLoader as Loader
except ImportError:
    from yaml import Loader

import monasca_agent.common.singleton as singleton

DEFAULT_CONFIG_FILE = '/etc/monasca/agent/agent.conf'
DEFAULT_LOG_DIR = '/var/log/monasca/agent'
LOGGING_MAX_BYTES = 5 * 1024 * 1024

log = logging.getLogger(__name__)


class Config(object):
    # Make this a singleton class so we don't get the config every time
    # the class is created
    __metaclass__ = singleton.Singleton

    def __init__(self, configFile=None):
        self._config = None
        if configFile is not None:
            self._configFile = configFile
        elif os.path.exists(DEFAULT_CONFIG_FILE):
            self._configFile = DEFAULT_CONFIG_FILE
        elif os.path.exists(os.getcwd() + '/agent.conf'):
            self._configFile = os.getcwd() + '/agent.conf'
        else:
            log.error('No config file found at {} nor in the working directory.'.format(DEFAULT_CONFIG_FILE))

        self._read_config()

    def get_config(self, sections='Main'):
        """Get the config info."""
        section_list = []
        if isinstance(sections, six.string_types):
            section_list.append(sections)
        elif isinstance(sections, list):
            section_list.extend(sections)
        else:
            log.error('Unknown section: {}'.format(str(sections)))
            return {}

        new_config = {}
        for section in section_list:
            new_config.update(self._config[section])

        return new_config

    def get_version(self):
        return pkg_resources.require("monasca-agent")[0].version

    def _read_config(self):
        """Read in the config file."""

        file_config = parser.SafeConfigParser()
        log.debug("Loading config file from {0}".format(self._configFile))
        file_config.readfp(self._skip_leading_wsp(open(self._configFile)))
        self._config = self._retrieve_sections(file_config)

        # Process and update any special case configuration
        self._parse_config()

    def _retrieve_sections(self, config):
        """Get the section values from the config file."""

        # Define default values for the possible config items
        the_config = {'Main': {'check_freq': 15,
                               'forwarder_url': 'http://localhost:17123',
                               'hostname': None,
                               'dimensions': None,
                               'listen_port': None,
                               'version': self.get_version(),
                               'additional_checksd': os.path.join(os.path.dirname(self._configFile), '/checks_d/'),
                               'system_metrics': None,
                               'ignore_filesystem_types': None,
                               'device_blacklist_re': None,
                               'limit_memory_consumption': None,
                               'skip_ssl_validation': False,
                               'watchdog': True,
                               'use_mount': False,
                               'autorestart': False,
                               'non_local_traffic': False},
                      'Api': {'is_enabled': False,
                              'url': '',
                              'project_name': '',
                              'project_id': '',
                              'project_domain_name': '',
                              'project_domain_id': '',
                              'ca_file': '',
                              'insecure': '',
                              'username': '',
                              'password': '',
                              'use_keystone': True,
                              'keystone_url': '',
                              'max_buffer_size': 1000,
                              'backlog_send_rate': 5},
                      'Statsd': {'recent_point_threshold': None,
                                 'monasca_statsd_interval': 20,
                                 'monasca_statsd_forward_host': None,
                                 'monasca_statsd_forward_port': 8125,
                                 'monasca_statsd_port': 8125},
                      'Logging': {'disable_file_logging': False,
                                  'log_level': None,
                                  'collector_log_file': DEFAULT_LOG_DIR + '/collector.log',
                                  'forwarder_log_file': DEFAULT_LOG_DIR + '/forwarder.log',
                                  'statsd_log_file': DEFAULT_LOG_DIR + '/statsd.log',
                                  'jmxfetch_log_file': DEFAULT_LOG_DIR + '/jmxfetch.log',
                                  'log_to_event_viewer': False,
                                  'log_to_syslog': False,
                                  'syslog_host': None,
                                  'syslog_port': None}}

        # Load values from configuration file into config file dictionary
        for section in config.sections():
            for option in config.options(section):
                try:
                    option_value = config.get(section, option)
                    if option_value == -1:
                        log.debug("Config option missing: {0}, using default value of {1}".format(option,
                                                                                                  the_config[section][
                                                                                                      option]))
                    else:
                        the_config[section][option] = option_value

                except Exception:
                    log.error("exception on %s!" % option)

        return the_config

    def _skip_leading_wsp(self, file):
        """Works on a file, returns a file-like object"""
        return cstringio.StringIO("\n".join(map(string.strip, file.readlines())))

    def _parse_config(self):
        # Parse_dimensions
        if self._config['Main']['dimensions'] is not None:
            # parse comma separated dimensions into a dimension list
            dimensions = {}
            try:
                dim_list = [dim.split(':') for dim in self._config['Main']['dimensions'].split(',')]
                dimensions.update(dict((key.strip(), value.strip()) for key, value in dim_list))
            except ValueError:
                log.info("Unable to process dimensions.")
                dimensions = {}
            self._config['Main']['dimensions'] = dimensions

        # Parse system metrics
        if self._config['Main']['system_metrics']:
            # parse comma separated system metrics into a metrics list
            try:
                metrics_list = [x.strip() for x in self._config['Main']['system_metrics'].split(',')]
            except ValueError:
                log.info("Unable to process system_metrics.")
                metrics_list = []
            self._config['Main']['system_metrics'] = metrics_list

        # Parse device blacklist regular expression
        try:
            filter_device_re = self._config['Main']['device_blacklist_re']
            if filter_device_re:
                self._config['Main']['device_blacklist_re'] = re.compile(filter_device_re)
        except re.error as err:
            log.error('Error processing regular expression {0}'.format(filter_device_re))

        # Parse file system types
        if self._config['Main']['ignore_filesystem_types']:
            # parse comma separated file system types to ignore list
            try:
                file_system_list = [x.strip() for x in self._config['Main']['ignore_filesystem_types'].split(',')]
            except ValueError:
                log.info("Unable to process ignore_filesystem_types.")
                file_system_list = []
            self._config['Main']['ignore_filesystem_types'] = file_system_list

    def get_confd_path(self):
        path = os.path.join(os.path.dirname(self._configFile), 'conf.d')
        if os.path.exists(path):
            return path
        raise PathNotFound(path)

    def check_yaml(self, conf_path):
        f = open(conf_path)
        try:
            check_config = yaml.load(f.read(), Loader=Loader)
            assert 'init_config' in check_config, "No 'init_config' section found"
            assert 'instances' in check_config, "No 'instances' section found"

            valid_instances = True
            if check_config['instances'] is None or not isinstance(check_config['instances'], list):
                valid_instances = False
            else:
                for i in check_config['instances']:
                    if not isinstance(i, dict):
                        valid_instances = False
                        break
            if not valid_instances:
                raise Exception(
                    'You need to have at least one instance defined in the YAML file for this check')
            else:
                return check_config
        finally:
            f.close()


def main():
    configuration = Config()
    config = configuration.get_config()
    api_config = configuration.get_config('Api')
    print "Main Configuration: \n {}".format(config)
    print "\nApi Configuration: \n {}".format(api_config)


if __name__ == "__main__":
    main()



""" Aggregation classes used by the collector and statsd to batch messages sent to the forwarder.
"""
import logging
from time import time

import monasca_agent.common.metrics as metrics_pkg


log = logging.getLogger(__name__)

# This is used to ensure that metrics with a timestamp older than
# RECENT_POINT_THRESHOLD_DEFAULT seconds (or the value passed in to
# the MetricsAggregator constructor) get discarded rather than being
# input into the incorrect bucket. Currently, the MetricsAggregator
# does not support submitting values for the past, and all values get
# submitted for the timestamp passed into the flush() function.
RECENT_POINT_THRESHOLD_DEFAULT = 3600


class MetricsAggregator(object):

    """
    A metric aggregator class.
    """

    def __init__(self, hostname, recent_point_threshold=None):
        self.events = []
        self.total_count = 0
        self.count = 0
        self.event_count = 0
        self.hostname = hostname

        recent_point_threshold = recent_point_threshold or RECENT_POINT_THRESHOLD_DEFAULT
        self.recent_point_threshold = int(recent_point_threshold)
        self.num_discarded_old_points = 0

        self.metrics = {}
        self.metric_type_to_class = {
            'g': metrics_pkg.Gauge,
            'c': metrics_pkg.Counter,
            'h': metrics_pkg.Histogram,
            'ms': metrics_pkg.Histogram,
            's': metrics_pkg.Set,
            'r': metrics_pkg.Rate,
        }

    def decrement(self, name, value=-1, dimensions=None, delegated_tenant=None,
                  hostname=None, device_name=None):
        self.submit_metric(name, value, 'c', dimensions, delegated_tenant,
                           hostname, device_name)

    def event(
            self,
            title,
            text,
            date_happened=None,
            alert_type=None,
            aggregation_key=None,
            source_type_name=None,
            priority=None,
            dimensions=None,
            hostname=None):
        event = {
            'msg_title': title,
            'msg_text': text,
            }
        if date_happened is not None:
            event['timestamp'] = date_happened
        else:
            event['timestamp'] = int(time())
        if alert_type is not None:
            event['alert_type'] = alert_type
        if aggregation_key is not None:
            event['aggregation_key'] = aggregation_key
        if source_type_name is not None:
            event['source_type_name'] = source_type_name
        if priority is not None:
            event['priority'] = priority
        if dimensions is not None:
            event['dimensions'] = dimensions
        if hostname is not None:
            event['host'] = hostname
        else:
            event['host'] = self.hostname

        self.events.append(event)

    def flush(self):
        timestamp = time()

        # Flush samples.  The individual metrics reset their internal samples
        # when required
        metrics = []
        for context, metric in self.metrics.items():
            metrics.extend(metric.flush(timestamp))

        # Log a warning regarding metrics with old timestamps being submitted
        if self.num_discarded_old_points > 0:
            log.warn('{0} points were discarded as a result of having an old timestamp'.format(
                     self.num_discarded_old_points))
            self.num_discarded_old_points = 0

        # Save some stats.
        log.debug("received {0} payloads since last flush".format(self.count))
        self.total_count += self.count
        self.count = 0
        return metrics

    def flush_events(self):
        events = self.events
        self.events = []

        self.total_count += self.event_count
        self.event_count = 0

        log.debug("Received {0} events since last flush".format(len(events)))

        return events

    @staticmethod
    def formatter(metric, value, timestamp, dimensions, hostname,
                  delegated_tenant=None, device_name=None, metric_type=None):
        """ Formats metrics, put them into a Measurement class
            (metric, timestamp, value, {"dimensions": {"name1": "value1", "name2": "value2"}, ...})
            dimensions should be a dictionary
        """
        if hostname:
            dimensions.update({'hostname': hostname})
        if device_name:
            dimensions.update({'device': device_name})

        return metrics_pkg.Measurement(metric,
                                       int(timestamp),
                                       value,
                                       dimensions,
                                       delegated_tenant)

    def gauge(self, name, value, dimensions=None, delegated_tenant=None,
              hostname=None, device_name=None, timestamp=None):
        self.submit_metric(name, value, 'g', dimensions, delegated_tenant,
                           hostname, device_name, timestamp)

    def histogram(self, name, value, dimensions=None, delegated_tenant=None,
                  hostname=None, device_name=None):
        self.submit_metric(name, value, 'h', dimensions, delegated_tenant,
                           hostname, device_name)

    def increment(self, name, value=1, dimensions=None, delegated_tenant=None,
                  hostname=None, device_name=None):
        self.submit_metric(name, value, 'c', dimensions, delegated_tenant,
                           hostname, device_name)

    def packets_per_second(self, interval):
        if interval == 0:
            return 0
        return round(float(self.count) / interval, 2)

    def rate(self, name, value, dimensions=None, delegated_tenant=None,
             hostname=None, device_name=None):
        self.submit_metric(name, value, 'r', dimensions, delegated_tenant,
                           hostname, device_name)

    def set(self, name, value, dimensions=None, delegated_tenant=None,
            hostname=None, device_name=None):
        self.submit_metric(name, value, 's', dimensions, delegated_tenant,
                           hostname, device_name)

    def submit_metric(self, name, value, mtype, dimensions=None,
                      delegated_tenant=None, hostname=None, device_name=None,
                      timestamp=None, sample_rate=1):

        context = (name, tuple(dimensions.items()),
                   delegated_tenant, hostname, device_name)

        if context not in self.metrics:
            metric_class = self.metric_type_to_class[mtype]
            self.metrics[context] = metric_class(self.formatter,
                                                 name,
                                                 dimensions,
                                                 hostname or self.hostname,
                                                 device_name,
                                                 delegated_tenant)
        cur_time = time()
        if timestamp is not None:
            if cur_time - int(timestamp) > self.recent_point_threshold:
                log.debug("Discarding {0} - ts = {1}, current ts = {2} ".format(name, timestamp, cur_time))
                self.num_discarded_old_points += 1
                return
        else:
            timestamp = cur_time
        self.metrics[context].sample(value, sample_rate, timestamp)
